/*
 * Copyright (c) 2018, University of Kaiserslautern
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors:
 *    Felipe S. Prado
 *    Matthias Jung
 */

#ifndef IARBITER_H_
#define IARBITER_H_

#include <tlm.h>
#include <systemc.h>
#include <iostream>
#include <vector>
#include <queue>
#include <tlm_utils/multi_passthrough_target_socket.h>
#include <tlm_utils/multi_passthrough_initiator_socket.h>
#include "../controller/core/configuration/ConfigurationLoader.h"

using namespace std;
using namespace tlm;

struct IArbiter: public sc_module {
public:
    tlm_utils::multi_passthrough_target_socket<IArbiter> tSocket;
    tlm_utils::multi_passthrough_initiator_socket<IArbiter> iSocket;

    SC_CTOR(IArbiter)
    {
        // One or more devices can accesss all the memory units through the arbiter.
        // Devices' initiator sockets are bound to arbiter's target sockets.
        // As soon the arbiter receives a request in any of its target sockets it should treat and forward it to the proper memory channel.
        tSocket.register_nb_transport_fw(this, &IArbiter::nb_transport_fw);

        // The arbiter communicates with one or more memory unity through one or more sockets (one or more memory channels).
        // Each of the arbiter's initiator sockets is bound to a memory controller's target socket.
        // Anytime an transaction comes from a memory unity to the arbiter the "bw" callback is called.
        iSocket.register_nb_transport_bw(this, &IArbiter::nb_transport_bw);

        tSocket.register_transport_dbg(this, &IArbiter::transport_dbg);
    }

    virtual void setTlmRecorder(TlmRecorder *recorder) = 0;
    virtual bool isOutputBufferFull(unsigned int initiatorSocket) = 0;
    virtual void incrementNumberOfOutputBufferTransactions(unsigned int
                                                           initiatorSocket) = 0;

protected:
    // Initiated by initiator side
    // This function is called when an arbiter's target socket receives a transaction from a device
    virtual tlm_sync_enum nb_transport_fw(int id, tlm_generic_payload &payload,
                                          tlm_phase &phase, sc_time &fwDelay) = 0;

    // Initiated by dram side
    // This function is called when an arbiter's initiator socket receives a transaction from a memory controller
    virtual tlm_sync_enum nb_transport_bw(int channelId,
                                          tlm_generic_payload &payload, tlm_phase &phase, sc_time &bwDelay) = 0;

    virtual unsigned int transport_dbg(int /*id*/,
                                       tlm::tlm_generic_payload &trans) = 0;

    void printDebugMessage(std::string message)
    {
        DebugManager::getInstance().printDebugMessage(this->name(), message);
    }

    void appendDramExtension(int socketId, tlm_generic_payload &payload)
    {
        // Append Generation Extension
        GenerationExtension *genExtension = new GenerationExtension(clkAlign(
                                                                        sc_time_stamp(), Configuration::getInstance().ControllerClk));
        payload.set_auto_extension(genExtension);

        unsigned int burstlength = payload.get_streaming_width();
        DecodedAddress decodedAddress = xmlAddressDecoder::getInstance().decodeAddress(
                                            payload.get_address());
        // Check the valid range of decodedAddress
        if (addressIsValid(decodedAddress)) {
            DramExtension *extension = new DramExtension(Thread(socketId + 1),
                                                         Channel(decodedAddress.channel), Bank(decodedAddress.bank),
                                                         BankGroup(decodedAddress.bankgroup), Row(decodedAddress.row),
                                                         Column(decodedAddress.column), burstlength);
            payload.set_auto_extension(extension);
        } else {
            SC_REPORT_FATAL("Arbiter", "Decoded Address are not inside the valid range");
        }
    }

    bool addressIsValid(DecodedAddress  &decodedAddress)
    {
        if (decodedAddress.channel >=
                xmlAddressDecoder::getInstance().amount["channel"]) {
            return false;
        }
        if (decodedAddress.bank >= xmlAddressDecoder::getInstance().amount["bank"]) {
            return false;
        }
        if (decodedAddress.bankgroup >
                xmlAddressDecoder::getInstance().amount["bankgroup"]) {
            return false;
        }
        if (decodedAddress.column >=
                xmlAddressDecoder::getInstance().amount["column"]) {
            return false;
        }
        if (decodedAddress.row >= xmlAddressDecoder::getInstance().amount["row"]) {
            return false;
        }
        return true;
    }
};

#endif /* IARBITER_H_ */
